#include <stdio.h>
#include "pico/stdlib.h"
#include "hardware/spi.h"

// TEST 0 = release , TEST 1 = TEST
#define TEST 0

#if TEST
#define __MAIN__
#endif

#include "keyboard.h"

// SPI Defines
// We are going to use SPI 0, and allocate it to the following GPIO pins
// Pins can be changed, see the GPIO function select table in the datasheet for information on GPIO assignments
#define SPI_PORT spi0
#define PIN_MISO 16
#define PIN_CS   17
#define PIN_SCK  18
#define PIN_MOSI 19

#define MCP23S17_CONFIG	0
#define MCP23S17_PULLUP	1
#define MCP23S17_INPUT	2
#define MCP23S17_OUTPUT	3
#define MCP23S17_PORTA	0
#define MCP23S17_PORTB	1

uint8_t regs[4][2] = {
	{ 0x00, 0x01 },  // CONFIG
	{ 0x0C, 0x0D },  // PULLUP
	{ 0x12, 0x13 },  // INPUT
	{ 0x14, 0x15 },  // OUTPUT
};

uint16_t keycode[3][12][8] = {
    {
    //Bit 7,6,  5,  4,  3,  2,  1,  0
    {0,0,0,0,0,0,0,2}, // SHIFT
    {0,0,0,0,0,0,0,1}, // ON/OFF (CTRL)
    //{"",   "",  "(x)","a","q", "z", "1","(x)"},
    {   0,    0,    0,  0x04,0x14,0x1D,0x1E,0},
    //{"f6"," ",  "y", "g", "6", "b", "f5","(x)"},
    {  0x3F,0x2c, 0x1C,0x0A,0x23,0x05,0x3E, 0},
    //{"f3","ひら", "t","f", "5", "v", "f4","(x)"},
    { 0x3C, 0x3F, 0x17,0x09,0x22,0x19,0x3D,  0},
    //{"f2","英", "r","d", "4", "c",  "3","(x)"},
    { 0x3B, 0x40,0x15,0x07,0x21,0x06,0x20,  0},
    //{"f1","全", "e", "s", "w", "x", "2", "(x)"},
    { 0x3A, 0x35, 0x08,0x16,0x1A,0x1B,0x1F,  0},
    //{"¥", "^", "7", "8", "9", "0", "=", "(x)"},
    {  0x89,0x2E,0x24,0x25,0x26,0x27,0x2D,  0},
    //{"backspace","[",  "u","i", "o", "p","@",  "(x)"},
    {  0x2A,       0x30,0x18,0x0C,0x12,0x13,0x2F,  0},
    //{"del",":", "h", "j", "k", "l", ";", "(x)"},
    {  0x4C, 0x34,0x0B,0x0D,0x0E,0x0F,0x33,  0},
    //{"]","\\", "(x)", "m", ",", ".", "/", "(x)"},
    { 0x32,0x89,   0  ,0x10,0x36,0x37,0x38,  0},
    //{"\n","↑", "n","変換","⇚","⇒","⇓","(x)"},
    {  0x28,0x52,0x11,0x8A,0x50,0x4F,0x51, 0},
    },
    { // *********************************** CTRL ****************
    {0,0,0,0,0,0,0,2}, // SHIFT
    {0,0,0,0,0,0,0,1}, // ON/OFF (CTRL)
    //{"",     "", "(x)","A",   "Q",  "Z",   "!",    "(x)"}, 
    { 0x0100,0x0100, 0, 0x0104,0x0114,0x011D, 0x011E,  0},
    //{"F6",   " ",    "Y",   "G",   "&",   "B",   "F5",  "(x)"},
    { 0x0100,0x012C, 0x011C,0x010A,0x0100,0x0105,0x0100,   0},
    //{"F3", "ROMAZI", "T",  "F",   "%",   "V",   "F4","(x)"},
    { 0x0100, 0x0100,0x0117,0x0109,0x0100,0x0119,0x0100,0},
    //{"F2", "かな",   "R",   "D",   "$",   "C",   "#",  "(x)"},
    { 0x0100, 0x0100,0x0115,0x0107,0x0100,0x0106,0x0100,  0},
    //{"F1","半",    "E",   "S",   "W",   "X",   "\"",  "(x)"},
    { 0x0100,0x0100,0x0108,0x0116,0x011A,0x011B, 0x0100,  0},
    //{"|",  "_",   "\"",   "(",   ")",   "~",   "-",   "(x)"},
    { 0x0100,0x0100,0x0100,0x0100,0x0100,0x0100,0x0100,   0},
    //{"backspace","{",  "U",    "I",   "O",   "P",  "`",   "(x)"},
    {   0x012A,  0x0100,0x0118,0x010C,0x0112, 0x0113,0x0100,  0},
    //{"del","*",    "H",   "J",   "K",   "L",   "+",   "(x)"},
    { 0x014C,0x0100,0x010B,0x010D,0x010E,0x010F,0x0100,   0},
    //{"}",  "_",   "(x)","M",  "<",   ">",   "?",   "(x)"},
    { 0x0100,0x0100,  0, 0x0110,0x0100,0x0100,0x0100,  0},
    //{"\n", "↑",    "N",   "変換",  "⇚",  "⇒",  "⇓",  "(x)"},
    { 0x0128,0x0152, 0x0111,0x0100,0x0150,0x014F,0x0151,  0},
    },
    { // *********************************** SHIFT ****************
    {0,0,0,0,0,0,0,2}, // SHIFT
    {0,0,0,0,0,0,0,1}, // ON/OFF (CTRL)
    //{"",   "",   "(x)","A",   "Q",  "Z",   "!",   "(x)"},
    { 0x0200,0x0200, 0, 0x0204,0x0214,0x021D,0x021E,  0},
    //{"F6", " ",    "Y",   "G",   "&",   "B",   "F5",  "(x)"},
    { 0x0045,0x022C, 0x021C,0x020A,0x0223,0x0205,0x0044,  0},
    //{"F3", "ROMAZI", "T",  "F",   "%",   "V",   "F4","(x)"},
    { 0x0042, 0x0200 ,0x0217,0x0209,0x0222,0x0219,0x0043,0},
    //{"F2", "かな", "R",    "D",   "$",   "C",   "#",  "(x)"},
    { 0x0041,0x0200,0x0215,0x0207,0x0221,0x0206,0x0220,  0},
    //{"F1","半",    "E",   "S",   "W",   "X",   "\"",  "(x)"},
    { 0x0040,0x0235, 0x0208,0x0216,0x021A,0x021B,0x021F,  0},
    //{"|",  "_",   "\"",  "(",   ")",   "~",   "-",   "(x)"},
    { 0x0289,0x022E,0x0224,0x0225,0x0226,0x0227,0x022D,  0},
    //{"backspace","{",  "U",    "I",   "O",   "P",    "`",   "(x)"},
    {   0x002A,   0x0230, 0x0218,0x020C,0x0212,0x0213,0x022F,  0},
    //{"del","*",    "H",   "J",   "K",   "L",   "+",   "(x)"},
    { 0x024C,0x0234, 0x020B,0x020D,0x020E,0x020F,0x0233,  0},
    //{"}",  "_",   "(x)","M",  "<",   ">",   "?",   "(x)"},
    { 0x0232,0x0289,  0, 0x0210,0x0236,0x0237,0x0238,  0},
    //{"\n", "↑",    "N",   "変換",   "⇚",  "⇒",   "⇓",  "(x)"},
    { 0x0228,0x0252, 0x0211, 0x0200,0x0250,0x024F,0x0251,  0},
    },
};
uint16_t kbuf[2][12];
uint16_t kcode = 0;
uint16_t ckin = 0;

void write_config(int addr, int portab, uint8_t value) {
    uint8_t writeData[] = {0x40 |(addr<<1), regs[MCP23S17_CONFIG][portab], value};
    gpio_put(PIN_CS, 0);
    spi_write_blocking(SPI_PORT, writeData, 3);
    gpio_put(PIN_CS, 1);
}

void write_pullup(int addr,int portab, uint8_t value) {
    uint8_t writeData[] = {0x40 |(addr<<1), regs[MCP23S17_PULLUP][portab], value};
    gpio_put(PIN_CS, 0);
    spi_write_blocking(SPI_PORT, writeData, 3);
    gpio_put(PIN_CS, 1);
}

uint8_t read_config(int addr,int portab) {
    uint8_t writeData[] = {0x41 |(addr<<1), regs[MCP23S17_CONFIG][portab], 0x00};
    uint8_t buff[3];
    gpio_put(PIN_CS, 0);
    spi_write_read_blocking(SPI_PORT, writeData, buff, 3);
    gpio_put(PIN_CS, 1);
    return buff[2];
}

void write_output(int addr,int portab, uint8_t value) {
    uint8_t writeData[] = {0x40 |(addr<<1), regs[MCP23S17_OUTPUT][portab], value};
    gpio_put(PIN_CS, 0);
    spi_write_blocking(SPI_PORT, writeData, 3);
    gpio_put(PIN_CS, 1);
}

uint8_t read_output(int addr,int portab) {
    uint8_t writeData[] = {0x41 |(addr<<1), regs[MCP23S17_OUTPUT][portab], 0x00};
    uint8_t buff[3];
    gpio_put(PIN_CS, 0);
    spi_write_read_blocking(SPI_PORT, writeData, buff, 3);
    gpio_put(PIN_CS, 1);
    return buff[2];
}

uint8_t read_input(int  addr,int portab) {
    uint8_t writeData[] = {0x41 |(addr<<1), regs[MCP23S17_INPUT][portab], 0x00};
    uint8_t buff[3];
    gpio_put(PIN_CS, 0);
    spi_write_read_blocking(SPI_PORT, writeData, buff, 3);
    gpio_put(PIN_CS, 1);
    return buff[2];
}

void init_MCP23S17() {
    // SPI initialisation. This example will use SPI at 1MHz.
    spi_init(SPI_PORT, 1000*1000);
    gpio_set_function(PIN_MISO, GPIO_FUNC_SPI);
    gpio_set_function(PIN_CS,   GPIO_FUNC_SIO);
    gpio_set_function(PIN_SCK,  GPIO_FUNC_SPI);
    gpio_set_function(PIN_MOSI, GPIO_FUNC_SPI);
    
    // Chip select is active-low, so we'll initialise it to a driven-high state
    gpio_set_dir(PIN_CS, GPIO_OUT);
    gpio_put(PIN_CS, 1);

    write_config(0,MCP23S17_PORTA,0xFF);  // MCP23S17-1 Port-A Mode Input
    write_config(0,MCP23S17_PORTB,0x00);  // MCP23S17-1 Port-B Mode Output
    write_pullup(0,MCP23S17_PORTA,0xFF);  // MCP23S17-1 Port-A PULLUP
    write_config(4,MCP23S17_PORTA,0x00);  // MCP23S17-2 Port-A Mode Output 
    write_config(4,MCP23S17_PORTB,0x00);  // MCP23S17-2 Port-B Mode Output
    write_output(4,MCP23S17_PORTA,0xFF);  // MCP23S17-2 Port-A Bit7-0 High
    write_output(4,MCP23S17_PORTB,0xff);  // MCP23S17-2 Port-B Bit7-0 High
}

// return 0xFFFF = None Key
//        0xFFFF != Key Action
//        0x0000 = Key Up
//
uint16_t keyin() {
    uint16_t shift = 0;
    uint8_t k;
    uint16_t kin = 0xFFFF;
    uint8_t i;

    write_output(4,MCP23S17_PORTA,0xFF);  // MCP23S17-2 Port-A Bit7-0 High
    write_output(4,MCP23S17_PORTB,0xFF);  // MCP23S17-2 Port-B Bit7-0 High
    for ( i = 0; i < 12; i++) {
        int a = i / 8;
        int bit = i % 8;
        write_output(4,MCP23S17_PORTA+a, (1<<bit) ^ 0xFF); // MCP23S17-2 Port-A,B Bit 0 -> 11 is Low at 1bit
        k = read_input(0, MCP23S17_PORTA);   // MCP23S17-1 Port-A 8 Bits Input
        kbuf[0][i] = kbuf[1][i] | k;         // 8 Bits Chata absorption
        kbuf[1][i] = k;                      // save 8 Bits
        write_output(4,MCP23S17_PORTA+a,0xFF); // MCP23S17-2 Port-A,B Bits is All High
    }
    if ((kbuf[0][1] & 0x01) == 0x00) {
         shift |= 2;   // ON SHIFT KEY
    }
    if ((kbuf[0][0] & 0x01) == 0x00) {
         shift |= 1;   // ON CTRL KEY
    }
    for ( i = 0; i < 10; i++) {
        if (kbuf[0][i+2] != 0xFF ) break;    // if On Key ?
    }
    for (int b = 0; i < 10 && b < 8; b++) {
        if ((kbuf[0][i+2] & (0x80>>b)) == 0x00) { // Serch 1 Bit is Low ?
            kin = keycode[shift % 3][i+2][b];     // Hit USB Keyboard Code & SHIFT,CTRL Bit
            break;
        }
    }
    if ( kin == 0xffff && shift != 0 ) kin = shift << 8; // Set SHIFT, CTRL Bits

    if ( kcode != kin ) {          // if Change latest key push ?
        ckin = 0;                  //  clear counter
        if ( kin == 0xFFFF ) {     // if current No key ?
            kcode = kin;           //  save Key code
            kin = 0x0000;        //  latest KEY Push and current No Key ==> KEY UP
        } else {
            kcode = kin;           // set current Key code
        }
    } else {
        if ( ckin < 5 ) {          // Adjust Update Key-in Timing
            if ( kin != 0xFFFF) ckin++;
            kin = 0xFFFF;
        }
    }

    return kin;
}

#ifdef __MAIN__
int main()
{
    uint8_t bt = 4;
    uint16_t kb;

    stdio_init_all();
    putchar('>');
    getchar();
    printf("**** START ****\n");
    init_MCP23S17();
    while (1) {
        //kb = read_input(0, MCP23S17_PORTA);
        //printf("In:%02X\n", kb ^ 0xFF);
        //if ( kb != 0xFF ) {
        //    char buf[30];
        //    printf("In:%02X\n", kb ^ 0xFF);
        //    putchar('>');
        //    gets(buf);
        //    while ( buf[0] != 0) {
        //        bt = atoi(buf);
        //        printf("input:%d\n", bt);
        //        putchar('?');
        //        gets(buf);
        //        if ( buf[0] == 0) break;
        //    }
        //    write_output(4,MCP23S17_PORTA,bt ^ 0xFF);
        //}
        //printf("OUT:%04X\n",keyin());
        kb = keyin();
        if (kb != 0xFFFF) printf("In:%04X\n", kb);
        sleep_ms(100);
    }
    return 0;
}
#endif
